import time
import typing as t
import googleapiclient.discovery
from google.oauth2 import service_account
from mlcube_gcp.gcp_client.operation import Operation


class Service(object):
    """
    https://cloud.google.com/compute/docs/tutorials/python-guide
    https://stackoverflow.com/questions/51303178/launch-gcp-instance-from-my-pc-using-python
    https://stackoverflow.com/questions/49444290/googleapiclient-authentication-using-personal-account
    """
    def __init__(self, project_id: t.Text, zone: t.Text, credentials: t.Optional[t.Text] = None) -> None:
        self.project_id: t.Text = project_id
        self.zone: t.Text = zone
        if isinstance(credentials, t.Dict) and 'file' in credentials:
            credentials = service_account.Credentials.from_service_account_file(
                credentials.get('file'),
                scopes=credentials.get('scopes', ['https://www.googleapis.com/auth/cloud-platform'])
            )
        else:
            credentials = None
        self.service = googleapiclient.discovery.build('compute', 'v1', credentials=credentials)

    def list_instances(self) -> t.List:
        response = self.service.instances().list(project=self.project_id, zone=self.zone).execute()
        return response.get('items', [])

    def get_instance(self, name: t.Text) -> t.Optional[t.Dict]:
        for instance in self.list_instances():
            if instance.get('name', None) == name:
                return instance
        return None

    def start_instance(self, name: t.Text) -> t.Dict:
        return self.service.instances().start(project=self.project_id, zone=self.zone, instance=name).execute()

    def stop_instance(self, name: t.Text) -> t.Dict:
        return self.service.instances().stop(project=self.project_id, zone=self.zone, instance=name).execute()

    def delete_instance(self, name: t.Text) -> t.Dict:
        return self.service.instances().delete(project=self.project_id, zone=self.zone, instance=name).execute()

    def create_instance(self, **kwargs) -> t.Dict:
        """
        https://cloud.google.com/compute/docs/reference/rest/v1/instances/setMachineType
        Assumed: https://cloud.google.com/compute/docs/instances/adding-removing-ssh-keys#project-wide
        https://cloud.google.com/compute/docs/reference/rest/v1/instances/insert
        """
        name = kwargs.get('name', 'gcp-f1-micro')
        machine_type = kwargs.get('machine_type', 'f1-micro')
        image = kwargs.get('image', {'project': 'ubuntu-os-cloud', 'family': 'ubuntu-1804-lts'})
        disk_size_gb = kwargs.get('disk_size_gb', 20)

        image_response = self.service.images().getFromFamily(project=image['project'], family=image['family']).execute()
        config: t.Dict = {
            'name': name,
            'machine_type': f"zones/{self.zone}/machineTypes/{machine_type}",
            'disks': [{
                'boot': True,
                'autoDelete': True,
                'initializeParams': {
                    'sourceImage': image_response['selfLink'],
                    'diskSizeGb': disk_size_gb
                }
            }],
            'networkInterfaces': [{
                'network': 'global/networks/default',
                'accessConfigs': [{
                    "kind": "compute#accessConfig",
                    'type': 'ONE_TO_ONE_NAT',
                    'name': 'External NAT',
                    "networkTier": "PREMIUM",
                }]
            }],
            'serviceAccounts': [{
                'email': 'default',
                'scopes': [
                    'https://www.googleapis.com/auth/devstorage.read_write',
                    'https://www.googleapis.com/auth/logging.write',
                    "https://www.googleapis.com/auth/monitoring.write",
                    "https://www.googleapis.com/auth/servicecontrol",
                    "https://www.googleapis.com/auth/service.management.readonly",
                    "https://www.googleapis.com/auth/trace.append"
                ]
            }],
        }
        return self.service.instances().insert(project=self.project_id, zone=self.zone, body=config).execute()

    def wait_for_operation(self, operation: t.Union[t.Text, t.Dict, Operation], retry_pause: float = 5):
        if isinstance(operation, t.Dict):
            operation = operation.get('name', None)
        elif isinstance(operation, Operation):
            operation = operation.name
        while True:
            result = self.service.zoneOperations().get(
                project=self.project_id, zone=self.zone, operation=operation
            ).execute()
            if result['status'] == 'DONE':
                if 'error' in result:
                    raise Exception(result['error'])
                return result
            time.sleep(retry_pause)



import typing as t


class Status(object):
    RUNNING = 'RUNNING'
    STOPPING = 'STOPPING'
    TERMINATED = 'TERMINATED'


class Instance(object):
    """ Just a wrapper around GCP instance dict v1. """

    def __init__(self, instance: t.Optional[t.Dict]) -> None:
        self.instance: t.Dict = instance if isinstance(instance, dict) else {}

    @property
    def name(self) -> t.Optional[t.Text]:
        return self.instance.get('name', None)

    @property
    def id(self) -> t.Optional[t.Text]:
        return self.instance.get('id', None)

    @property
    def status(self) -> t.Optional[t.Text]:
        return self.instance.get('status', None)

    @property
    def public_ip(self) -> t.Optional[t.Text]:
        for interface in self.instance.get('networkInterfaces', None) or []:
            for access_config in interface.get('accessConfigs', None) or []:
                if access_config.get('name', '') == 'External NAT':
                    return access_config.get('natIP', None)
        return None

    def __str__(self) -> t.Text:
        return f"Instance(name={self.name}, id={self.id}, status={self.status}, public_ip={self.public_ip})"



import click
import typing as t
from mlcube_gcp.gcp_client.service import Service
from mlcube_gcp.gcp_client.instance import (Instance, Status)


@click.group(name='mlcube_gcp_client')
def cli():
    """ """
    pass


@cli.command(name='list', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
def list_instances(project_id: t.Text, zone: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    print("Instances:")
    for instance in service.list_instances():
        print(f"\t{Instance(instance)}")


@cli.command(name='status', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
@click.argument('name', required=True, type=str)
def instance_status(project_id: t.Text, zone: t.Text, name: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    print(Instance(service.get_instance(name=name)))


@cli.command(name='start', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
@click.argument('name', required=True, type=str)
def instance_start(project_id: t.Text, zone: t.Text, name: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    instance = Instance(service.get_instance(name=name))
    if instance.name is None:
        print(f"Instance '{name}' does not exist.")
    elif instance.status == Status.RUNNING:
        print(f"Instance '{name}' is already running: {instance}")
    else:
        service.wait_for_operation(service.start_instance(instance.name))
        print(f"Instance '{name}' has started: {Instance(service.get_instance(name))}")


@cli.command(name='stop', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
@click.argument('name', required=True, type=str)
def instance_stop(project_id: t.Text, zone: t.Text, name: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    instance = Instance(service.get_instance(name=name))
    if instance.name is None:
        print(f"Instance '{name}' does not exist.")
    elif instance.status != Status.RUNNING:
        print(f"Instance '{name}' is not running: {instance}")
    else:
        service.wait_for_operation(service.stop_instance(instance.name))
        print(f"Instance '{name}' has stopped: {Instance(service.get_instance(name))}")


@cli.command(name='create', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
def instance_create(project_id: t.Text, zone: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    service.wait_for_operation(service.create_instance())


@cli.command(name='delete', help='List instances.')
@click.option('--project-id', required=True, type=str, help='Project identifier.')
@click.option('--zone', required=True, type=str, help='Zone.')
@click.argument('name', required=True, type=str)
def instance_delete(project_id: t.Text, zone: t.Text, name: t.Text) -> None:
    service = Service(project_id=project_id, zone=zone)
    instance = Instance(service.get_instance(name=name))
    if instance.name is None:
        print(f"Instance '{name}' does not exist.")
    else:
        service.wait_for_operation(service.delete_instance(instance.name))
        print(f"Instance '{name}' has been deleted: {Instance(service.get_instance(name))}")


if __name__ == '__main__':
    cli()



import typing as t


class Operation(object):
    def __int__(self, operation: t.Dict) -> None:
        self.operation: t.Dict = operation

    @property
    def id(self) -> t.Optional[t.Text]:
        return self.operation.get('id', None)

    @property
    def name(self) -> t.Optional[t.Text]:
        return self.operation.get('name', None)

    @property
    def type(self) -> t.Optional[t.Text]:
        return self.operation.get('operationType', None)

    @property
    def progress(self) -> t.Optional[t.Text]:
        return self.operation.get('progress', None)



import os
import logging
import typing as t
from pathlib import Path
from omegaconf import (DictConfig, OmegaConf)
from mlcube.validate import Validate
from mlcube_ssh.ssh_run import Shell
from ssh_config.client import (SSHConfig, Host)
from mlcube.runner import (RunnerConfig, Runner)
from mlcube.errors import ExecutionError
from mlcube_gcp.gcp_client.instance import Instance as GCPInstance, Status as GCPInstanceStatus
from mlcube_gcp.gcp_client.service import Service


logger = logging.getLogger(__name__)


class Config(RunnerConfig):

    DEFAULT = OmegaConf.create({
        'runner': 'gcp',

        'gcp': {
            'project_id': '',
            'zone': '',
            'credentials': ''
        },
        'instance': {
            'name': '',
            'machine_type': '',
            'disk_size_gb': ''
        },
        'platform': ''
    })

    @staticmethod
    def validate(mlcube: DictConfig) -> None:
        Validate(mlcube.runner, 'runner')\
            .check_unknown_keys(Config.DEFAULT.keys())\
            .check_values(['gcp', 'instance'], DictConfig)\
            .check_values(['platform'], str, blanks=False)
        Validate(mlcube.runner.gcp, 'runner.gcp')\
            .check_unknown_keys(['project_id', 'zone', 'credentials'])\
            .check_values(['project_id', 'zone'], str, blanks=False) \
            .not_none(['credentials'])
        Validate(mlcube.runner.instance, 'runner.instance')\
            .check_unknown_keys(['name', 'machine_type', 'disk_size_gb'])\
            .not_none(['name', 'machine_type', 'disk_size_gb']) \
            .check_values(['name', 'machine_type'], str, blanks=False)


class GCPRun(Runner):

    CONFIG = Config

    def __init__(self, mlcube: t.Union[DictConfig, t.Dict], task: t.Text) -> None:
        super().__init__(mlcube, task)

    def configure(self) -> None:
        """  """
        gcp: DictConfig = self.mlcube.runner

        # Check that SSH is configured.
        # TODO: (Sergey) why am I doing it here (copy-past bug)?
        ssh_config_file = os.path.join(Path.home(), '.ssh', 'mlcube')
        try:
            ssh_config = SSHConfig.load(ssh_config_file)
            gcp_host: Host = ssh_config.get(gcp.instance.name)
        except KeyError:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                f"SSH configuration file ({ssh_config_file}) does not provide connection details for GCP instance "
                f"(name={gcp.instance.name}). Most likely this error has occurred due to implementation error - "
                "please, contact MLCube developers."
            )
        # TODO: I can try to add this info on the fly assuming standard paths. Need to figure out the user name.
        if gcp_host.get('User', None) is None or gcp_host.get('IdentityFile', None) is None:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                f"SSH configuration file ({ssh_config_file}) provides connection details for GCP instance "
                f"(name={gcp.instance.name}), but these details do not include information about `User` "
                "and/or `IdentifyFile`."
            )

        # Connect to GCP
        logger.info("Connecting to GCP ...")
        try:
            service = Service(project_id=gcp.gcp.project_id, zone=gcp.gcp.zone, credentials=gcp.gcp.credentials)
        except Exception as err:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                "The error most like is associated with either reading credentials, or connecting using google API ("
                f"project_id={gcp.gcp.project_id}, zone={gcp.gcp.zone}, credentials={gcp.gcp.credentials}). See "
                "context for more details.",
                error=str(err),
                gcp_info={'project_id': gcp.gcp.project_id, 'zone': gcp.gcp.zone, 'credentials': gcp.gcp.credentials}
            )

        # Figure out if an instance needs to be created
        try:
            instance = GCPInstance(service.get_instance(gcp.instance.name))
            if instance.name is None:
                print("Creating GCP instance ...")
                service.wait_for_operation(
                    service.create_instance(name=gcp.instance.name, machine_type=gcp.instance.machine_type,
                                            disk_size_gb=gcp.instance.disk_size_gb)
                )
                instance = GCPInstance(service.get_instance(gcp.instance.name))

            # Check its running status
            if instance.status != GCPInstanceStatus.RUNNING:
                print("Starting GCP instance ...")
                service.wait_for_operation(service.start_instance(instance.name))
                instance = GCPInstance(service.get_instance(gcp.instance.name))
        except Exception as err:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                "Failed to create or connect to remote GCP instance. See context for more details.",
                error=str(err),
                gcp_instance_info={
                    'name': gcp.instance.name, 'machine_type': gcp.instance.machine_type,
                    'disk_size_gb': gcp.instance.disk_size_gb
                }
            )

        # Make sure SSH mlcube is up-to-date
        if gcp_host.get('HostName', None) != instance.public_ip:
            print(f"Updating SSH mlcube (prev={gcp_host.get('HostName')}, new={instance.public_ip}, "
                  f"file={ssh_config_file})")
            ssh_config.update(instance.name, {'HostName': instance.public_ip})
            ssh_config.write(ssh_config_file)
            # TODO: clean '.ssh/known_hosts'.

        # Configure remote instance. This is specific for docker-based images now.
        try:
            Shell.ssh(
                gcp.instance.name,
                'sudo snap install docker && sudo addgroup --system docker && sudo adduser ${USER} docker && '
                'sudo snap disable docker && sudo snap enable docker && '
                'sudo apt update && yes | sudo apt install python3-pip virtualenv && sudo apt clean'
            )
        except ExecutionError as err:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                "Failed to install system packages on a remote instance. See context for more details.",
                error=str(err)
            )

        # Remote GCP instance has been configured
        print(instance)

        # Should be as simple as invoking SSH configure.
        try:
            Shell.run(f"mlcube configure --mlcube={self.mlcube.root} --platform={gcp.platform}")
        except ExecutionError as err:
            raise ExecutionError.mlcube_configure_error(
                self.__class__.__name__,
                f"Error occurred while running mlcube configure with GCP platform (platform={gcp.platform}). See "
                "context for more details.",
                error=str(err)
            )

    def run(self) -> None:
        gcp: DictConfig = self.mlcube.runner
        try:
            Shell.run(f"mlcube run --mlcube={self.mlcube.root} --platform={gcp.platform} --task={self.task}")
        except ExecutionError as err:
            raise ExecutionError.mlcube_run_error(
                self.__class__.__name__,
                f"Error occurred while running MLCube task (platform={gcp.platform}, task={self.task}).",
                **err.context
            )
